# -*- coding: utf-8 -*-
"""
Created on Sat Mar 19 19:58:53 2022

@author: Petr Tomasek, Institute of Mathematics, Faculty of Mechanical Engineering, Brno University of Technology, Czech Republic
"""
# 
# Numerical solver for IVP to Caputo FDE
# 
# y'(x)=f(x,y(x)),  y(xinit)=yinit
#
# by Forward Euler method, Backward Euler method, Theta method
#
#
#import fnGraph as fnG
import numpy as np
import math
from scipy.optimize import fsolve
import multiprocessing as mp
from joblib import Parallel, delayed


def fnFrwEulMethod(f,alpha,tinit,yinit,tfinal,IC,h):
  t=np.arange(IC[0],tfinal+h,h) # Mesh initialization
  y=np.empty(len(t));
  fi=np.empty(len(t)); # storing of f(t[i],y[i]) init
  b=np.empty(len(t)); # b weights init
  y[0]=IC[1];
  for n in range(0,len(t)-1):
      fi[n]=f(t[n],y[n]);
      P1=0;
      for j in range(0,n+1):
          b[j]=(pow(n+1-j,alpha)-pow(n-j,alpha))/math.gamma(alpha+1); #weight b
          P1+=b[j]*fi[j];
      y[n+1]=yinit+pow(h,alpha)*P1;   
#  fnG.gr1(t, y)
  return y;
  
  
def fnBkwEulMethod(f,alpha,tinit,yinit,tfinal,IC,h): 
  t=np.arange(IC[0],tfinal+h,h) # Mesh initialization
  y=np.empty(len(t));
  fi=np.empty(len(t)); # storing of f(t[i],y[i]) init
  b=np.empty(len(t)); # b weights init
  y[0]=IC[1];
  for n in range(0,len(t)-1):
      fi[n]=f(t[n],y[n]);
      P1=0;
      for j in range(0,n):
          b[j]=(pow(n+1-j,alpha)-pow(n-j,alpha))/math.gamma(alpha+1); #weight b
          P1+=b[j]*fi[j+1];
      b[n]=(pow(1,alpha)-pow(0,alpha))/math.gamma(alpha+1); #weight b
      #y[n+1]=yinit+pow(h,alpha)*P1;   
      def bweulstep(z): #RHS of DE
          return(yinit+pow(h,alpha)*(P1+b[n]*f(t[n+1],z))-z);
      yguess=y[n];
      y[n+1]=fsolve(bweulstep,yguess)
#  fnG.gr1(t, y)
  return y;

def fnThetaMethod(f,alpha,tinit,yinit,tfinal,IC,h,Theta): 
  t=np.arange(IC[0],tfinal+h,h) # Mesh initialization
  y=np.empty(len(t));
  fi=np.empty(len(t)); # storing of f(t[i],y[i]) init
  b=np.empty(len(t)); # b weights init
  y[0]=IC[1];
  for n in range(0,len(t)-1):
      fi[n]=f(t[n],y[n]);
      P1=0;
      for j in range(0,n):
          b[j]=(pow(n+1-j,alpha)-pow(n-j,alpha))/math.gamma(alpha+1); #weight b
          P1+=b[j]*(Theta*fi[j]+(1-Theta)*fi[j+1]);
      b[n]=(pow(1,alpha)-pow(0,alpha))/math.gamma(alpha+1); #weight b
      #y[n+1]=yinit+pow(h,alpha)*P1;   
      def thstep(z): #RHS of DE
          return(yinit+pow(h,alpha)*(P1+b[n]*(Theta*fi[n]+(1-Theta)*f(t[n+1],z)))-z);
      yguess=y[n];
      y[n+1]=fsolve(thstep,yguess)
#  fnG.gr1(t, y)
  return y;



def fnFrwEulMethodParallel(f,alpha,tinit,yinit,tfinal,IC,h):
  print(f"Number of cpu: {mp.cpu_count()}")
  
    

  n_cpu=mp.cpu_count()
  #pool = mp.Pool(processes=n_cpu)
  t=np.arange(IC[0],tfinal+h,h) # Mesh initialization
  y=np.empty(len(t));
  fi=np.empty(len(t)); # storing of f(t[i],y[i]) init
  b=np.empty(len(t)); # b weights init
  y[0]=IC[1];
  
  def P1comp(j):
      b[j]=(pow(n+1-j,alpha)-pow(n-j,alpha))/math.gamma(alpha+1); #weight b
      return b[j]*fi[j]
  
  for n in range(0,math.floor(4*len(t)/5)):
      fi[n]=f(t[n],y[n]);
      P1=0;
      for j in range(0,n+1):
          b[j]=(pow(n+1-j,alpha)-pow(n-j,alpha))/math.gamma(alpha+1); #weight b
          P1+=b[j]*fi[j];
      y[n+1]=yinit+pow(h,alpha)*P1;     
  
  for n in range(math.floor(4*len(t)/5),len(t)-1):
      fi[n]=f(t[n],y[n]);
      P1arr=[];
      
      
      #for j in range(0,n+1):
      #    b[j]=(pow(n+1-j,alpha)-pow(n-j,alpha))/math.gamma(alpha+1); #weight b
      #    P1arr[j]=b[j]*fi[j];
      P1arr=Parallel(n_jobs=2,backend="loky",verbose=0)(delayed(P1comp)(i) for i in range(0,n+1))
      #P1arr=Parallel(n_jobs=4, prefer='threads', require='sharedmem')(delayed(P1comp)(s) for s in range(0,n+1))
     # P1arr=[pool.map(P1comp, range(0,n+1))]
      P1=sum(P1arr)  
      y[n+1]=yinit+pow(h,alpha)*P1;   
#  fnG.gr1(t, y)
  return y;

